/***************************************************************************

Cobra Command:
  Video hardware looks like early version of Bad Dudes, we appear to have
  2 playfields and 1 text layer.  Each playfield has 2 scroll registers and
  4 control bytes:

  Register 1: MSB set is reverse screen (playfield 1 only), always seems to be 2
  Register 2: Unknown, seems to be always 0
  Register 3: Unknown, 3 in foreground, 0 in background (so transparency setting?)
  Register 4: Probably playfield shape, only 2 * 2 supported at moment (1)

  Sprite hardware appears to be the same as Bad Dudes.
  256 colours, palette generated by ram.

The Real Ghostbusters:
  1 playfield, same as above, with rowscroll
  1024 colours from 2 proms.
  Sprite hardware close to above, there are some unused (unknown) bits per sprite.

Super Real Darwin:
  1 playfield, x-scroll only
  Closer to earlier Darwin board than above games.

Last Mission:
  Sprite hardware same as Karnov.

***************************************************************************/

#include "driver.h"
#include "vidhrdw/generic.h"

static unsigned char *palette_ram,*pf_video,*pf_dirty,*oscar_shared_mem;
static int scroll1[4],scroll2[4],pf1_attr[8],pf2_attr[8];
static struct osd_bitmap *pf1_bitmap,*pf2_bitmap;
unsigned char *dec8_row;

extern int ghost_prot;

/***************************************************************************

  Convert the color PROMs into a more useable format.

  Real Ghostbusters has two 1024x8 palette PROM.
  I don't know the exact values of the resistors between the RAM and the
  RGB output. I assumed these values (the same as Commando)

  bit 7 -- 220 ohm resistor  -- GREEN
        -- 470 ohm resistor  -- GREEN
        -- 1  kohm resistor  -- GREEN
        -- 2.2kohm resistor  -- GREEN
        -- 220 ohm resistor  -- RED
        -- 470 ohm resistor  -- RED
        -- 1  kohm resistor  -- RED
  bit 0 -- 2.2kohm resistor  -- RED

  bit 7 -- unused
        -- unused
        -- unused
        -- unused
        -- 220 ohm resistor  -- BLUE
        -- 470 ohm resistor  -- BLUE
        -- 1  kohm resistor  -- BLUE
  bit 0 -- 2.2kohm resistor  -- BLUE

***************************************************************************/
void ghostb_vh_convert_color_prom(unsigned char *palette, unsigned short *colortable,const unsigned char *color_prom)
{
	int i;

	for (i = 0;i < Machine->drv->total_colors;i++)
	{
		int bit0,bit1,bit2,bit3;

		bit0 = (color_prom[0] >> 0) & 0x01;
		bit1 = (color_prom[0] >> 1) & 0x01;
		bit2 = (color_prom[0] >> 2) & 0x01;
		bit3 = (color_prom[0] >> 3) & 0x01;
		*(palette++) = 0x0e * bit0 + 0x1f * bit1 + 0x43 * bit2 + 0x8f * bit3;
		bit0 = (color_prom[0] >> 4) & 0x01;
		bit1 = (color_prom[0] >> 5) & 0x01;
		bit2 = (color_prom[0] >> 6) & 0x01;
		bit3 = (color_prom[0] >> 7) & 0x01;
		*(palette++) = 0x0e * bit0 + 0x1f * bit1 + 0x43 * bit2 + 0x8f * bit3;
		bit0 = (color_prom[Machine->drv->total_colors] >> 0) & 0x01;
		bit1 = (color_prom[Machine->drv->total_colors] >> 1) & 0x01;
		bit2 = (color_prom[Machine->drv->total_colors] >> 2) & 0x01;
		bit3 = (color_prom[Machine->drv->total_colors] >> 3) & 0x01;
		*(palette++) = 0x0e * bit0 + 0x1f * bit1 + 0x43 * bit2 + 0x8f * bit3;

		color_prom++;
	}
}

void dec8_pf1_w(int offset, int data)
{
	switch (offset)
	{
  	case 0:
    case 2:
    case 4:
    case 6:
      pf1_attr[offset]=data;
  }
}

void dec8_pf2_w(int offset, int data)
{
	switch (offset)
	{
		case 0:
		case 2:
		case 4:
		case 6:
			pf2_attr[offset]=data;
  }
}

void dec8_scroll1_w(int offset, int data)
{
	scroll1[offset]=data;
}

void dec8_scroll2_w(int offset, int data)
{
	scroll2[offset]=data;
}

void srdarwin_control_w(int offset, int data)
{
	int bankaddress;
	unsigned char *RAM = Machine->memory_region[Machine->drv->cpu[0].memory_region];

	switch (offset) {
    	case 1:
        	/* Hmm */
        	return;

      case 2:
               	return;

    	case 3: /* Top 3 bits - bank switch, bottom 4 - scroll MSB */
			bankaddress = 0x10000 + (data >> 5) * 0x4000;
			cpu_setbank(1,&RAM[bankaddress]);

         scroll2[0]=data&0xf;
         return;

        case 4:
        	scroll2[1]=data;
        	return;
    }

}

void lastmiss_control_w(int offset, int data)
{
	int bankaddress;
	unsigned char *RAM = Machine->memory_region[Machine->drv->cpu[0].memory_region];

   /* Bottom 4 bits - bank switch, Bits 4 & 5 - Scroll MSBs */
	 bankaddress = 0x10000 + (data & 0x0f) * 0x4000;
   cpu_setbank(1,&RAM[bankaddress]);

   scroll2[0]=(data>>5)&1;
   scroll2[2]=(data>>6)&1;

    scroll1[0]=data;
}

void lastmiss_scrollx_w(int offset, int data)
{
	scroll2[1]=data;
}

void lastmiss_scrolly_w(int offset, int data)
{
	scroll2[3]=data;
}


/******************************************************************************/

void dec8_vh_screenrefresh(struct osd_bitmap *bitmap, int full_refresh)
{
	int my,mx,offs,color,tile;
	int scrollx,scrolly;

	if (palette_recalc())
    	gm_memset(pf_dirty,1,0x800);

	/* Playfield 2 - Foreground */
	mx=-1; my=0;
	for (offs = 0x000;offs < 0x400; offs += 2) {
		mx++;
		if (mx==16) {mx=0; my++;}
		if (!pf_dirty[offs/2]) continue; else pf_dirty[offs/2]=0;
		tile=pf_video[offs+1]+(pf_video[offs]<<8);
		color = ((tile & 0xf000) >> 12);
        tile=tile&0xfff;

		drawgfx(pf2_bitmap,Machine->gfx[2],tile,
			color+12, 0,0, 16*mx,16*my,
		 	0,TRANSPARENCY_NONE,0);
	}

	mx=-1; my=0;
	for (offs = 0x400;offs < 0x800; offs += 2) {
	    mx++;
	    if (mx==16) {mx=0; my++;}
	    if (!pf_dirty[offs/2]) continue; else pf_dirty[offs/2]=0;
	    tile=pf_video[offs+1]+(pf_video[offs]<<8);
	    color = ((tile & 0xf000) >> 12);
	    tile=tile&0xfff;
		drawgfx(pf2_bitmap,Machine->gfx[2],tile,
			color+12, 0,0, (16*mx)+256,16*my,
		 	0,TRANSPARENCY_NONE,0);
	}

  /* Playfield 1 *
  my=-1; mx=0;
  for (offs = 0x800;offs < 0xc00; offs += 2) {
    my++;
    if (my==16) {my=0; mx++;}
    if (!pf_dirty[offs/2]) continue; else pf_dirty[offs/2]=0;
    tile=pf_video[offs+1]+(pf_video[offs]<<8);
    color = ((tile & 0xf000) >> 12);
    tile=tile&0xfff;
		drawgfx(pf1_bitmap,Machine->gfx[3],tile,
			color+4, 0,0, (16*mx)+256,16*my,
		 	0,TRANSPARENCY_NONE,0);
	}

  my=-1; mx=0;
  for (offs = 0xc00;offs < 0x1000; offs += 2) {
    my++;
    if (my==16) {my=0; mx++;}
    if (!pf_dirty[offs/2]) continue; else pf_dirty[offs/2]=0;
    tile=pf_video[offs+1]+(pf_video[offs]<<8);
    color = ((tile & 0xf000) >> 12);
    tile=tile&0xfff;
		drawgfx(pf1_bitmap,Machine->gfx[3],tile,
			color+4, 0,0, 16*mx,(16*my)+256,
		 	0,TRANSPARENCY_NONE,0);
	}

  scrolly=-((scroll1[2]<<8)+scroll1[3]);
  scrollx=-((scroll1[0]<<8)+scroll1[1]);
  copyscrollbitmap(bitmap,pf1_bitmap,1,&scrollx,1,&scrolly,0,TRANSPARENCY_NONE,0);
          */

	scrolly=-((scroll2[2]<<8)+scroll2[3]);
	scrollx=-((scroll2[0]<<8)+scroll2[1]);
	copyscrollbitmap(bitmap,pf2_bitmap,1,&scrollx,1,&scrolly,0,TRANSPARENCY_NONE,0);

	/* Sprites */
	for (offs = 0;offs < 0x800;offs += 8)
	{
		int x,y,sprite,colour,multi,fx,fy,inc;

		y =spriteram[offs+1]+(spriteram[offs]<<8);
		if ((y&0x8000) == 0) continue;
		x = spriteram[offs+5]+(spriteram[offs+4]<<8);
		colour = ((x & 0x3000) >> 12)+4;
		fx = y & 0x2000;
		fy = y & 0x4000;
		multi = (1 << ((y & 0x1800) >> 11)) - 1;	/* 1x, 2x, 4x, 8x height */

											/* multi = 0   1   3   7 */
		sprite = spriteram[offs+3]+(spriteram[offs+2]<<8);
		sprite &= 0x0fff;

		x = x & 0x01ff;
		y = y & 0x01ff;
		if (x >= 256) x -= 512;
		if (y >= 256) y -= 512;
		x = 240 - x;
		y = 240 - y;

		sprite &= ~multi;
		if (fy)
			inc = -1;
		else
		{
			sprite += multi;
			inc = 1;
		}

		while (multi >= 0)
		{
			drawgfx(bitmap,Machine->gfx[1],
					sprite - multi * inc,
					colour,
					fx,fy,
					x,y - 16 * multi,
					&Machine->drv->visible_area,TRANSPARENCY_PEN,0);
			multi--;
		}
	}

	/* Draw character tiles */
	for (offs = 0x800 - 2;offs >= 0;offs -= 2) {
		tile=videoram[offs+1]+((videoram[offs]&0xF)<<8);
		if (!tile) continue;
		color=(videoram[offs]&0x70)>>4;
		mx = (offs/2) % 32;
	   	my = (offs/2) / 32;

        drawgfx(bitmap,Machine->gfx[0],
				tile,color,0,0,8*mx,8*my,
				&Machine->drv->visible_area,TRANSPARENCY_PEN,0);
   }

}

void ghostb_vh_screenrefresh(struct osd_bitmap *bitmap, int full_refresh)
{
	int my,mx,offs,color,tile;
	int scrollx,scrolly,fx,fy;

	/* Playfield */
	mx=-1; my=0;
	for (offs = 0x000;offs < 0x400; offs += 2) {
		mx++;
    	if (mx==16) {mx=0; my++;}
		if (!pf_dirty[offs/2]) continue; else pf_dirty[offs/2]=0;

    	tile=pf_video[offs+1]+(pf_video[offs]<<8);
    	color = ((tile & 0xf000) >> 12);
    	tile=tile&0xfff;
		drawgfx(pf2_bitmap,Machine->gfx[2],tile,
			color, 0, 0, 16*mx,16*my,
		 	0,TRANSPARENCY_NONE,0);
	}

	mx=-1; my=0;
	for (offs = 0x400;offs < 0x800; offs += 2) {
    	mx++;
    	if (mx==16) {mx=0; my++;}
    	if (!pf_dirty[offs/2]) continue; else pf_dirty[offs/2]=0;

	    tile=pf_video[offs+1]+(pf_video[offs]<<8);
	    color = ((tile & 0xf000) >> 12);
	    tile=tile&0xfff;
       drawgfx(pf2_bitmap,Machine->gfx[2],tile,
			color, 0, 0, (16*mx)+256,16*my,
		 	0,TRANSPARENCY_NONE,0);
	}

    /* Rowscroll */
    if (pf2_attr[0]&0x4) {
    	int rscroll[256];

		scrolly=-((scroll2[2]<<8)+scroll2[3]);

		for (offs = 0;offs < 512;offs+=2)
			rscroll[offs/2] = -( (dec8_row[offs]<<8) + dec8_row[offs+1] ) -((scroll2[0]<<8)+scroll2[1]);

		copyscrollbitmap(bitmap,pf2_bitmap,256,rscroll,1,&scrolly,&Machine->drv->visible_area,TRANSPARENCY_NONE,0);
    } else {
		scrolly=-((scroll2[2]<<8)+scroll2[3]);
		scrollx=-((scroll2[0]<<8)+scroll2[1]);
		copyscrollbitmap(bitmap,pf2_bitmap,1,&scrollx,1,&scrolly,&Machine->drv->visible_area,TRANSPARENCY_NONE,0);
    }

	/* Sprites */
	for (offs = 0;offs < 0x800;offs += 8)
	{
		int x,y,sprite,colour,multi,inc;
		y =spriteram[offs+1]+(spriteram[offs]<<8);
		if ((y&0x8000) == 0) continue;

		x = spriteram[offs+5]+(spriteram[offs+4]<<8);

		fx = spriteram[offs+3] & 0x4;
		fy = spriteram[offs+3] & 0x2;
		multi = (1 << ((y & 0x1800) >> 11)) - 1;	/* 1x, 2x, 4x, 8x height */

											/* multi = 0   1   3   7 */

		sprite = spriteram[offs+7]+(spriteram[offs+6]<<8);
		colour = sprite >> 12;

		sprite &= 0x0fff;

		x = x & 0x01ff;
		y = y & 0x01ff;

		if (x >= 256) x -= 512;
		if (y >= 256) y -= 512;
		x = 240 - x;
		y = 240 - y;

		sprite &= ~multi;

		if (fy)
			inc = -1;
		else
		{
			sprite += multi;
			inc = 1;
		}

		while (multi >= 0)
		{
			drawgfx(bitmap,Machine->gfx[1],
					sprite - multi * inc,
					colour,
					fx,fy,
					x,y - 16 * multi,
					&Machine->drv->visible_area,TRANSPARENCY_PEN,0);
			multi--;
		}
	}

	/* Draw character tiles */
	for (offs = 0x800 - 2;offs >= 0;offs -= 2) {
		tile=videoram[offs+1]+((videoram[offs]&0xF)<<8);
		if (!tile) continue;
		color=(videoram[offs]&0xc)>>2;
		mx = (offs/2) % 32;
		my = (offs/2) / 32;

		drawgfx(bitmap,Machine->gfx[0],
				tile,color,0,0,8*mx,8*my,
				&Machine->drv->visible_area,TRANSPARENCY_PEN,0);
	}

}

void srdarwin_vh_screenrefresh(struct osd_bitmap *bitmap, int full_refresh)
{
	int my,mx,offs,color,tile;
	int scrollx,scrolly;
    int bank;

	/* Playfield */
	mx=-1; my=0;
	for (offs = 0x000;offs < 0x400; offs += 2) {
		mx++;
    	if (mx==32) {mx=0; my++;}
		if (!pf_dirty[offs/2]) continue; else pf_dirty[offs/2]=0;

    	tile=pf_video[offs+1]+(pf_video[offs]<<8);
    	color=((tile & 0xf000) >> 12);
    	tile=tile&0xfff;
        bank=tile/0x100;
        tile-=(bank*0x100);

		drawgfx(pf2_bitmap,Machine->gfx[bank+2],tile,
			color, 0,0, 16*mx,16*my,
		 	0,TRANSPARENCY_NONE,0);
	}

	scrolly=0;
	scrollx=-((scroll2[0]<<8)+scroll2[1]);
	copyscrollbitmap(bitmap,pf2_bitmap,1,&scrollx,1,&scrolly,0,TRANSPARENCY_NONE,0);

	/* Sprites */
	for (offs = 0;offs < 0x200;offs += 4)
	{
		int multi,fx,fy,sx,sy,code;

   		if (spriteram[offs] == 0xf8) continue;

		sx = 240 - spriteram[offs+2];
		if (sx < -7) sx += 256;
		sy = 240 - spriteram[offs];
		code = spriteram[offs+3] + (256 * ((spriteram[offs+1]>>5) & 0x07));
		color = (spriteram[offs+1]>>3)&7;
        multi = spriteram[offs+1] & 0x80;
        fy=0; 
        fx=0; 

        if (multi) sy=sy-16;
    	drawgfx(bitmap,Machine->gfx[1],
        		code,
				color,
				fx,fy,
				sx,sy,
				&Machine->drv->visible_area,TRANSPARENCY_PEN,0);
        if (multi)
    		drawgfx(bitmap,Machine->gfx[1],
				code+1,
				color,
				fx,fy,
				sx,sy+16,
				&Machine->drv->visible_area,TRANSPARENCY_PEN,0);


	}

	/* Draw character tiles */
   	mx=-1; my=0;
	for (offs = 0;offs < 0x800; offs += 1) {
		mx++;
    	if (mx==32) {mx=0; my++;}
        tile=videoram[offs];
		if (!tile) continue;
        color=0;      
       	drawgfx(bitmap,Machine->gfx[0],
				tile,color,0,0,8*mx,8*my,
				&Machine->drv->visible_area,TRANSPARENCY_PEN,0);
    }

}

/******************************************************************************/

void lastmiss_vh_screenrefresh(struct osd_bitmap *bitmap, int full_refresh)
{
	int my,mx,offs,color,tile;
	int scrollx,scrolly;

	if (palette_recalc())
    	gm_memset(pf_dirty,1,0x800);

	/* Playfield */
	mx=-1; my=0;
	for (offs = 0x000;offs < 0x200; offs += 2) {
		mx++;
		if (mx==16) {mx=0; my++;}
		if (!pf_dirty[offs/2]) continue; else pf_dirty[offs/2]=0;
		tile=pf_video[offs+1]+(pf_video[offs]<<8);
		color = ((tile & 0xf000) >> 12);
        tile=tile&0xfff;

		drawgfx(pf2_bitmap,Machine->gfx[2],tile,
			color, 0,0, 16*mx,16*my,
		 	0,TRANSPARENCY_NONE,0);
	}

    mx=-1; my=0;
	for (offs = 0x200;offs < 0x400; offs += 2) {
		mx++;
		if (mx==16) {mx=0; my++;}
		if (!pf_dirty[offs/2]) continue; else pf_dirty[offs/2]=0;
		tile=pf_video[offs+1]+(pf_video[offs]<<8);
		color = ((tile & 0xf000) >> 12);
        tile=tile&0xfff;

		drawgfx(pf2_bitmap,Machine->gfx[2],tile,
			color, 0,0, (16*mx)+0x100,16*my,
		 	0,TRANSPARENCY_NONE,0);
	}

	mx=-1; my=0;
	for (offs = 0x400;offs < 0x600; offs += 2) {
	    mx++;
	    if (mx==16) {mx=0; my++;}
	    if (!pf_dirty[offs/2]) continue; else pf_dirty[offs/2]=0;
	    tile=pf_video[offs+1]+(pf_video[offs]<<8);
	    color = ((tile & 0xf000) >> 12);
	    tile=tile&0xfff;
		drawgfx(pf2_bitmap,Machine->gfx[2],tile,
			color, 0,0, 16*mx,(16*my)+256,
		 	0,TRANSPARENCY_NONE,0);
	}

	mx=-1; my=0;
	for (offs = 0x600;offs < 0x800; offs += 2) {
	    mx++;
	    if (mx==16) {mx=0; my++;}
	    if (!pf_dirty[offs/2]) continue; else pf_dirty[offs/2]=0;
	    tile=pf_video[offs+1]+(pf_video[offs]<<8);
	    color = ((tile & 0xf000) >> 12);
	    tile=tile&0xfff;
		drawgfx(pf2_bitmap,Machine->gfx[2],tile,
			color, 0,0, (16*mx)+256,(16*my)+0x100,
		 	0,TRANSPARENCY_NONE,0);
	}

	scrollx=-(scroll2[0]<<8)-scroll2[1];
	scrolly=-(scroll2[2]<<8)-scroll2[3];
	copyscrollbitmap(bitmap,pf2_bitmap,1,&scrollx,1,&scrolly,0,TRANSPARENCY_NONE,0);

	/* Sprites */
	for (offs = 0;offs < 0x800;offs += 8)
	{
		int x,y,sprite,colour,multi,fx,fy,inc;

		y =spriteram[offs+1]+(spriteram[offs]<<8);
		if ((y&0x8000) == 0) continue;
		x = spriteram[offs+5]+(spriteram[offs+4]<<8);
		colour = spriteram[offs+6] >> 5;
	 
	 
		multi = (1 << ((y & 0x1800) >> 11)) - 1;	/* 1x, 2x, 4x, 8x height */

											/* multi = 0   1   3   7 */
		sprite = spriteram[offs+7]+(spriteram[offs+6]<<8);
		sprite &= 0x0fff;

        fx=spriteram[offs+3];
        fy=fx&0x2;
        fx=fx&0x4;


		x = x & 0x01ff;
		y = y & 0x01ff;
		if (x >= 256) x -= 512;
		if (y >= 256) y -= 512;
		x = 240 - x;
		y = 240 - y;

		sprite &= ~multi;
		if (fy)
			inc = -1;
		else
		{
			sprite += multi;
			inc = 1;
		}

		while (multi >= 0)
		{
			drawgfx(bitmap,Machine->gfx[1],
					sprite - multi * inc,
					colour,
					fx,fy,
					x,y - 16 * multi,
					&Machine->drv->visible_area,TRANSPARENCY_PEN,0);
			multi--;
		}
	}


	/* Draw character tiles */
	for (offs = 0x800 - 2;offs >= 0;offs -= 2) {
		tile=videoram[offs+1]+((videoram[offs]&0xF)<<8);
		if (!tile) continue;
		color=videoram[offs]>>6;
		mx = (offs/2) % 32;
	   	my = (offs/2) / 32;

        drawgfx(bitmap,Machine->gfx[0],
				tile,color,0,0,8*mx,8*my,
				&Machine->drv->visible_area,TRANSPARENCY_PEN,0);
   }

}

/******************************************************************************/

int srdarwin_palette_rg_r(int offset)
{
	return palette_ram[offset];
}

int srdarwin_palette_b_r(int offset)
{
	return palette_ram[offset+0x400];
}

void srdarwin_palette_rg(int offset, int data)
{
	int r,g,b;

	palette_ram[offset]=data;

	r = 0x11 * ((palette_ram[offset] >> 0) & 0x0f);
	g = 0x11 * ((palette_ram[offset] >> 4) & 0x0f);
	b = 0x11 * ( palette_ram[offset+0x400] & 0x0f);

	palette_change_color(offset,r,g,b);
}

void srdarwin_palette_b(int offset, int data)
{
	int r,g,b;

	palette_ram[offset+0x400]=data;

	r = 0x11 * ((palette_ram[offset] >> 0) & 0x0f);
	g = 0x11 * ((palette_ram[offset] >> 4) & 0x0f);
	b = 0x11 * ( palette_ram[offset+0x400] & 0x0f);

	palette_change_color(offset,r,g,b);
}

/******************************************************************************/

int dec8_video_r(int offset)
{
	return pf_video[offset];
}

void dec8_video_w(int offset, int data)
{
	if (pf_video[offset]!=data)
	{
		pf_video[offset]=data;
		pf_dirty[offset/2] = 1;
	}
}

int oscar_share_r(int offset)
{
	return oscar_shared_mem[offset];
}

void oscar_share_w(int offset, int data)
{
	oscar_shared_mem[offset]=data;
}

int shackled_share_r(int offset)
{
	return oscar_shared_mem[offset+0x1000];
}

void shackled_share_w(int offset, int data)
{
	oscar_shared_mem[offset+0x1000]=data;
}

int oscar_share2_r(int offset)
{
	return oscar_shared_mem[offset+0x1000];
}

void oscar_share2_w(int offset, int data)
{
	unsigned char *RAM = Machine->memory_region[0];
	oscar_shared_mem[offset+0x1000]=data;

    /* The cpu's can execute program in the ram area */
	RAM[offset+0x1000]=data;
    RAM = Machine->memory_region[3];
	RAM[offset+0x1000]=data;
}

int dec8_vh_start (void)
{
	pf1_bitmap=osd_create_bitmap(512,512);
	pf2_bitmap=osd_create_bitmap(512,512);
	palette_ram=gm_zi_malloc(0x800);
	pf_video=gm_zi_malloc(0x1000);
	pf_dirty=gm_zi_malloc(0x800);
	oscar_shared_mem=gm_zi_malloc(0x2000);    
	gm_memset(pf_dirty,1,0x800);
	return 0;
}

void dec8_vh_stop (void)
{
	osd_free_bitmap(pf1_bitmap);
	osd_free_bitmap(pf2_bitmap);
	gm_free(palette_ram);
	gm_free(pf_video);
	gm_free(pf_dirty);
	gm_free(oscar_shared_mem);
	generic_vh_stop();
}



void oscar_vh_screenrefresh(struct osd_bitmap *bitmap, int full_refresh)
{
	int my,mx,offs,color,tile;
	int scrollx,scrolly;

	if (palette_recalc())
    	gm_memset(pf_dirty,1,0x800);

	/* Playfield */
	mx=-1; my=0;
	for (offs = 0x000;offs < 0x400; offs += 2) {
		mx++;
		if (mx==16) {mx=0; my++;}
		if (!pf_dirty[offs/2]) continue; else pf_dirty[offs/2]=0;
		tile=pf_video[offs+1]+(pf_video[offs]<<8);
		color = ((tile & 0xf000) >> 12);
        tile=tile&0xfff;

		drawgfx(pf2_bitmap,Machine->gfx[2],tile,
			color&7, 0,0, 16*mx,16*my,
		 	0,TRANSPARENCY_NONE,0);
	}

	mx=-1; my=0;
	for (offs = 0x400;offs < 0x800; offs += 2) {
	    mx++;
	    if (mx==16) {mx=0; my++;}
	    if (!pf_dirty[offs/2]) continue; else pf_dirty[offs/2]=0;
	    tile=pf_video[offs+1]+(pf_video[offs]<<8);
	    color = ((tile & 0xf000) >> 12);
	    tile=tile&0xfff;
		drawgfx(pf2_bitmap,Machine->gfx[2],tile,
			color&7, 0,0, (16*mx)+256,16*my,
		 	0,TRANSPARENCY_NONE,0);
	}

	scrolly=-((scroll2[2]<<8)+scroll2[3]);
	scrollx=-((scroll2[0]<<8)+scroll2[1]);
	copyscrollbitmap(bitmap,pf2_bitmap,1,&scrollx,1,&scrolly,0,TRANSPARENCY_NONE,0);

	/* Sprites */
	for (offs = 0;offs < 0x800;offs += 8)
	{
		int x,y,sprite,colour,multi,fx,fy,inc;

		y =spriteram[offs+1]+(spriteram[offs]<<8);
 		if ((y&0x8000) == 0) continue;
		x = spriteram[offs+5]+(spriteram[offs+4]<<8);
		colour = ((x & 0xf000) >> 12);
		fx = y & 0x2000;
		fy = y & 0x4000;
		multi = (1 << ((y & 0x1800) >> 11)) - 1;	/* 1x, 2x, 4x, 8x height */

											/* multi = 0   1   3   7 */
		sprite = spriteram[offs+3]+(spriteram[offs+2]<<8);
		sprite &= 0x0fff;

		x = x & 0x01ff;
		y = y & 0x01ff;
		if (x >= 256) x -= 512;
		if (y >= 256) y -= 512;
		x = 240 - x;
		y = 240 - y;

		sprite &= ~multi;
		if (fy)
			inc = -1;
		else
		{
			sprite += multi;
			inc = 1;
		}

		while (multi >= 0)
		{
			drawgfx(bitmap,Machine->gfx[1],
					sprite - multi * inc,
					colour,
					fx,fy,
					x,y - 16 * multi,
					&Machine->drv->visible_area,TRANSPARENCY_PEN,0);
			multi--;
		}
	}

	/* Draw character tiles */
	for (offs = 0x800 - 2;offs >= 0;offs -= 2) {
		tile=videoram[offs+1]+((videoram[offs]&0xF)<<8);
		if (!tile) continue;
		color=videoram[offs]>>6;
		mx = (offs/2) % 32;
	   	my = (offs/2) / 32;

        drawgfx(bitmap,Machine->gfx[0],
				tile,color,0,0,8*mx,8*my,
				&Machine->drv->visible_area,TRANSPARENCY_PEN,0);
   }

}
